
#include "Arduino.h"
#define NECHANT 128
#define SHIFT_ACCUM 25
#define GET_DATA 10
#define SET_DATA 11
#define DATA_0_SIZE 4 // fréquence, flottant 32 bits
#define DATA_1_SIZE 4 // amplitude, flottant 32 bits
uint8_t data_0[DATA_0_SIZE];
uint8_t data_1[DATA_1_SIZE];
bool data_0_ready = true;
bool data_0_request = false;
bool data_1_ready = true;
bool data_1_request = false;
float frequence = 10; // data_0
float amp = 0.0;
uint32_t period_pwm = 100; // en microsecondes (10 kHz)

uint32_t icr;
uint32_t table_onde_A[NECHANT];
uint32_t table_onde_B[NECHANT];
uint32_t indexA,indexB;
uint32_t accum1,accum2,increm;
uint16_t diviseur[6] = {0,1,8,64,256,1024};


// numérisation sonde Hall analogique
#define SAMPLE_PERIOD 1000
#define BUF_SIZE 500
volatile uint16_t buffer[2][BUF_SIZE];
volatile uint16_t indice_buf, nbuf;
volatile float B0_std;
volatile float B0_sum = 0;
volatile float B0_sum2 = 0;
#define NMOY 2000
volatile uint16_t imoy = 0;
#define DATA_2_SIZE 4 // B0_std, flottant 32 bits
uint8_t data_2[DATA_2_SIZE];
bool data_2_ready = true;
bool data_2_request = false;
#define DATA_3_SIZE 2000 // BUF_SIZE*4, buffer
uint8_t data_3[DATA_3_SIZE];
bool data_3_ready = true;
bool data_3_request = false;
#define DATA_4_SIZE 4 // Bstd cible, flottant 32 bits
uint8_t data_4[DATA_4_SIZE];
bool data_4_ready = true;
bool data_4_request = false;
float BstdCible;

//période de rotation
float period;
#define DATA_5_SIZE 4 // period, flottant 32 bits
uint8_t data_5[DATA_5_SIZE];
bool data_5_ready = true;
bool data_5_request = false;
unsigned long temps;

                          
                          
void init_pwm_timer1(uint32_t period) {
    char clockBits;
    cli();
    TCCR1A = 0;
    TCCR1A |= (1 << COM1A1); //Clear OCnA/OCnB/OCnC on compare match, set OCnA/OCnB/OCnC at BOTTOM (non-inverting mode)
    TCCR1A |= (1 << COM1B1);
    TCCR1A |= (1 << COM1C1);
    TCCR1B = 1 << WGM13; // phase and frequency correct pwm mode, top = ICR1
    int d = 1;
    icr = (F_CPU/1000000*period/2);
    while ((icr>0xFFFF)&&(d<6)) { // choix du diviseur d'horloge
        d++;
        icr = (F_CPU/1000000*period/2/diviseur[d]);
   } 
   clockBits = d;
   ICR1 = icr; // valeur maximale du compteur
   TIMSK1 = 1 << TOIE1; // overflow interrupt enable
   sei(); // activation des interruptions
   TCNT1 = 0; // mise à zéro du compteur
   TCCR1B |= clockBits; // déclenchement du compteur
}                         
                          
ISR(TIMER1_OVF_vect) { // Timer 1 Overflow interrupt
  accum1 += increm;
  accum2 += increm;
  indexA = accum1 >> SHIFT_ACCUM;
  OCR1A = table_onde_A[indexA];
  indexB = accum2 >> SHIFT_ACCUM;
  OCR1B = table_onde_B[indexB];
}                        
                          
void timer3_init(uint32_t period) {// timer pour l'ADC
  TCCR3A = 0;
    TCCR3B = 0;
    TCCR3B |= (1 << WGM12); // mode CTC avec OCR1A pour le maximum
    uint32_t top = (F_CPU/1000000*period);
    int clock = 1;
    while ((top>0xFFFF)&&(clock<5)) {
          clock++;
          top = (F_CPU/1000000*period/diviseur[clock]);
      }
    OCR3A = top; // période
    TIMSK3 = (1 << OCIE1A); // interruption lorsque TCNT3=OCR3A
    TCCR3B |= clock;
}
                              
                               
void adc_init(uint8_t prescaler) {
  ADCSRA = 0;
  ADCSRA |= (1 << ADEN); // enable ADC
  ADCSRA |= prescaler ;
  ADCSRB = 0;)
}                           
                               
ISR(TIMER3_COMPA_vect) { // gestionnaire d'interruption de l'ADC
    uint8_t v;
    uint16_t data;
    float B,Bm;
    
    ADMUX  = 0b01000000 | 0; // REFS0 et multiplex
    ADCSRA |= 0b01000000; // start ADC
    while (ADCSRA & 0b01000000);
    data = ADCL | (ADCH << 8);
    buffer[nbuf][indice_buf] = data;
    B = data;
    B0_sum += B;
    B0_sum2 += B*B;
     imoy += 1;
     if (imoy==NMOY) {
        Bm = B0_sum/NMOY;
        B0_std = sqrt(B0_sum2/NMOY-Bm*Bm)*0.1565;
        B0_sum = B0_sum2 = 0;
        data_2_ready = true;
        imoy = 0;
     }
     indice_buf += 1;
     if (indice_buf==BUF_SIZE) {
      indice_buf=0;
      if (nbuf==1) nbuf=0;
      else nbuf = 1;
      data_3_ready = true;
     }
}                                  
                                   
void set_sinus_table_A(float amp, float offset) {
  int i;
  float dt = 2*3.1415926/NECHANT;
  for(i=0; i<NECHANT; i++) {
    table_onde_A[i] = icr*0.5*(1.0+offset+amp*sin(i*dt));
  }  
}

void set_sinus_table_B(float amp, float offset) {
  int i;
  float dt = 2*3.1415926/NECHANT;
  for(i=0; i<NECHANT; i++) {
    table_onde_B[i] = icr*0.5*(1.0+offset+amp*sin(i*dt));
  }  
}                                  
                                   
void set_frequence() {
  increm = (uint32_t) (((float)(0xFFFFFFFF))*((float)(frequence)*1e-6*(float)(period_pwm))); // incrément de l'accumulateur de phase
}                                 
                                    
void compteur_tour() {
  unsigned long t = millis();
  if (t-temps > 200) {
      period = 1e-3*(t-temps);
      data_5_ready = true;
      temps = t;
  } 
}                                  
                                      
void setup() {
  Serial.begin(115200);
  Serial.setTimeout(0);
  char c = 0;
  Serial.write(c);
  c = 255;
  Serial.write(c);
  c = 0;
  Serial.write(c);
   pinMode(4,OUTPUT);
   digitalWrite(4,HIGH); // +5 V pour capteur
    pinMode(2,OUTPUT);
    digitalWrite(2,HIGH); // +5 V pour capteur
    pinMode(8,OUTPUT); 
    digitalWrite(8,HIGH); // commande ENABLE A
    pinMode(6,OUTPUT);
    digitalWrite(6,HIGH); // commande ENABLE B
    pinMode(11,OUTPUT); // PWM A, vers INA
    pinMode(12,OUTPUT); // PWM B, vers INB
    pinMode(3,INPUT); // sortie fourche optique
    for (int i=0; i<BUF_SIZE; i++) buffer[0][i] = buffer[1][i] = 0;
    cli();
    init_pwm_timer1(period_pwm);
    set_sinus_table_A(amp,0.0);
    set_sinus_table_B(amp,0.0);
    accum1 = 0;
    accum2 = ((uint32_t)(NECHANT * 0.25)) << SHIFT_ACCUM;
    set_frequence();
    indice_buf = 0;
    nbuf = 0;
    timer3_init(SAMPLE_PERIOD);
    adc_init(7);
    temps = millis();
    attachInterrupt(digitalPinToInterrupt(3), compteur_tour, FALLING);
    sei();
}                           
                               
void get_data() {
  char n;
  while (Serial.available()<1) {};
  n = Serial.read();
  if (n==0) {
    data_0_request = true;
    memcpy(data_0,&frequence,DATA_0_SIZE);
  }
  if (n==1) {
    data_1_request = true;
    memcpy(data_1,&amp,DATA_1_SIZE);
  }
  if (n==2) {
    data_2_request = true;
    memcpy(data_2,&B0_std,DATA_2_SIZE);
  }
  if (n==3) {
    data_3_request = true;
    if (nbuf==1) memcpy(data_3,&buffer[0],DATA_3_SIZE);
    else memcpy(data_3,&buffer[1],DATA_3_SIZE);
  }
  if (n==5) {
    data_5_request = true;
    memcpy(data_5,&period,DATA_5_SIZE);
  }
}                         
                                 
void send_data() {
  if ((data_0_ready)&&(data_0_request)) {
      //data_0_ready = false;
      data_0_request = false;
      Serial.write(data_0,DATA_0_SIZE);
  }
  if ((data_1_ready)&&(data_1_request)) {
      //data_1_ready = false;
      data_1_request = false;
      Serial.write(data_1,DATA_1_SIZE);
  }
  if ((data_2_ready)&&(data_2_request)) {
      data_2_ready = false;
      data_2_request = false;
      Serial.write(data_2,DATA_2_SIZE);
  }
  if ((data_3_ready)&&(data_3_request)) {
      data_3_ready = false;
      data_3_request = false;
      Serial.write(data_3,DATA_3_SIZE);
  }
  if ((data_5_ready)&&(data_5_request)) {
      //data_5_ready = false;
      data_5_request = false;
      Serial.write(data_5,DATA_5_SIZE);
  }
}                            
                                 
void read_serial() {
   char com;
   if (Serial.available()>0) {
        com = Serial.read();
        if (com==GET_DATA) get_data();
        else if (com==SET_DATA) set_data();
   }
}	                              
                                  
void loop() {
  read_serial();
  send_data();

}                              
                                  